#!/usr/bin/env python3
"""
Models that maps to Cloudformation functions.
"""


def replace_fn(node):
    """Iteratively replace all Fn/Ref in the node"""
    if isinstance(node, list):
        return [replace_fn(item) for item in node]
    if isinstance(node, dict):
        return {name: replace_fn(value) for name, value in node.items()}
    if isinstance(node, (str, int, float)):
        return node
    if isinstance(node, Ref):
        return node.render()
    if hasattr(Fn, node.__class__.__name__):
        return node.render()
    raise ValueError(f"Invalid value specified in the code: {node}")


class Ref:
    """Represents a ref function in Cloudformation."""

    # This is our DSL, it's a very thin wrapper around dictionary.
    # pylint: disable=R0903
    def __init__(self, target):
        """Creates a Ref node with a target."""
        self.target = target

    def render(self):
        """Render the node as a dictionary."""
        return {"Ref": self.target}


class Base64:
    """Fn::Base64 function."""

    # pylint: disable=R0903

    def __init__(self, value):
        self.value = value

    def render(self):
        """Render the node with Fn::Base64."""
        return {"Fn::Base64": replace_fn(self.value)}


class Cidr:
    """Fn::Cidr function."""

    # pylint: disable=R0903

    def __init__(self, ipblock, count, cidr_bits):
        self.ipblock = ipblock
        self.count = count
        self.cidr_bits = cidr_bits

    def render(self):
        """Render the node with Fn::Cidr."""
        return {
            "Fn::Cidr": [
                replace_fn(self.ipblock),
                replace_fn(self.count),
                replace_fn(self.cidr_bits),
            ]
        }


class And:
    """Fn::And function."""

    # pylint: disable=R0903

    def __init__(self, *args):
        self.conditions = list(args)

    def render(self):
        """Render the node with Fn::And."""
        return {"Fn::And": replace_fn(self.conditions)}


class Equals:
    """Fn::Equals function."""

    # pylint: disable=R0903

    def __init__(self, lhs, rhs):
        self.lhs = lhs
        self.rhs = rhs

    def render(self):
        """Render the node with Fn::Equals."""
        return {"Fn::Equals": [replace_fn(self.lhs), replace_fn(self.rhs)]}


class If:
    """Fn::If function."""

    # pylint: disable=R0903

    def __init__(self, condition, true_value, false_value):
        self.condition = condition
        self.true_value = true_value
        self.false_value = false_value

    def render(self):
        """Render the node with Fn::If."""
        return {
            "Fn::If": [
                replace_fn(self.condition),
                replace_fn(self.true_value),
                replace_fn(self.false_value),
            ]
        }


class Not:
    """Fn::Not function."""

    # pylint: disable=R0903

    def __init__(self, condition):
        self.condition = condition

    def render(self):
        """Render the node with Fn::Not."""
        return {"Fn::Not": [replace_fn(self.condition)]}


class Or:
    """Fn::Or function."""

    # pylint: disable=R0903

    def __init__(self, *args):
        self.conditions = list(args)

    def render(self):
        """Render the node with Fn::Or."""
        return {"Fn::Or": replace_fn(self.conditions)}


class FindInMap:
    """Fn::FindInMap function."""

    # pylint: disable=R0903

    def __init__(self, map_name, l1key, l2key):
        self.map_name = map_name
        self.l1key = l1key
        self.l2key = l2key

    def render(self):
        """Render the node with Fn::FindInMap."""
        return {
            "Fn::FindInMap": [
                replace_fn(self.map_name),
                replace_fn(self.l1key),
                replace_fn(self.l2key),
            ]
        }


class GetAtt:
    """Fn::GetAtt function."""

    # pylint: disable=R0903

    def __init__(self, logical_name, attr):
        self.logical_name = logical_name
        self.attr = attr

    def render(self):
        """Render the node with Fn::GetAtt."""
        return {"Fn::GetAtt": [replace_fn(self.logical_name), replace_fn(self.attr)]}


class GetAZs:
    """Fn::GetAZs function."""

    # pylint: disable=R0903

    def __init__(self, region):
        self.region = region

    def render(self):
        """Render the node with Fn::GetAZs."""
        return {"Fn::GetAZs": replace_fn(self.region)}


class ImportValue:
    """Fn::ImportValue function."""

    # pylint: disable=R0903

    def __init__(self, export):
        self.export = export

    def render(self):
        """Render the node with Fn::ImportValue."""
        return {"Fn::ImportValue": replace_fn(self.export)}


class Join:
    """Fn::Join function."""

    # pylint: disable=R0903

    def __init__(self, delimiter, elements):
        self.delimiter = delimiter
        self.elements = elements

    def render(self):
        """Render the node with Fn::Join."""
        return {"Fn::Join": [replace_fn(self.delimiter), replace_fn(self.elements)]}


class Select:
    """Fn::Select function."""

    # pylint: disable=R0903

    def __init__(self, index, elements):
        self.index = index
        self.elements = elements

    def render(self):
        """Render the node with Fn::Select."""
        return {"Fn::Select": [replace_fn(self.index), replace_fn(self.elements)]}


class Split:
    """Fn::Split function."""

    # pylint: disable=R0903

    def __init__(self, delimiter, target):
        self.delimiter = delimiter
        self.target = target

    def render(self):
        """Render the node with Fn::Split."""
        return {"Fn::Split": [replace_fn(self.delimiter), replace_fn(self.target)]}


class Sub:
    """Fn::Sub function."""

    # pylint: disable=R0903

    def __init__(self, target, mapping=None):
        if not isinstance(target, str):
            raise ValueError(
                f"The first argument of Fn::Sub must be string: `{target}`"
            )
        if mapping is None:
            self.mapping = {}

        self.target = target
        self.mapping = mapping

    def render(self):
        """Render the node with Fn::Sub."""
        if self.mapping:
            return {"Fn::Sub": [replace_fn(self.target), replace_fn(self.mapping)]}
        return {"Fn::Sub": replace_fn(self.target)}


class Transform:
    """Fn::Transform function."""

    # pylint: disable=R0903

    def __init__(self, construct):
        is_dict = isinstance(construct, dict)
        match_keys = set(construct.keys()) == {"Name", "Parameters"}

        if not is_dict or not match_keys:
            raise ValueError("Invalid Transform construct")

        self.construct = construct

    def render(self):
        """Render the node with Fn::Transform."""
        return {
            "Fn::Transform": {
                "Name": replace_fn(self.construct["Name"]),
                "Parameters": replace_fn(self.construct["Parameters"]),
            }
        }


class Fn:
    """
    This is a container for all functions.

    Rationale is instead of having to import all the functions,
    we just import Fn and use any function as Fn.FuncName
    """

    # pylint: disable=R0903

    Base64 = Base64
    Cidr = Cidr
    And = And
    Equals = Equals
    If = If
    Not = Not
    Or = Or
    FindInMap = FindInMap
    GetAtt = GetAtt
    GetAZs = GetAZs
    ImportValue = ImportValue
    Join = Join
    Select = Select
    Split = Split
    Sub = Sub
    Transform = Transform



#!/usr/bin/env python3
"""Model definitions in vapor."""
from .fn import replace_fn
from .utils import get_logger


logger = get_logger(__name__)


class ResourceBase(type):
    """Metaclass for all cfn resources."""

    def __new__(cls, name, bases, attrs):
        super_new = super().__new__

        # Only perform custom logic for the subclasses of Resource, but not Resource
        # itself.
        parents = [b for b in bases if isinstance(b, ResourceBase)]
        if not parents:
            return super_new(cls, name, bases, attrs)

        new_class = super_new(cls, name, bases, attrs)
        if "_module" in attrs:
            # We need to pass it down to the subclass
            setattr(new_class, "__module__", attrs["_module"].__name__)

        return new_class


class StackBase(type):
    """Metaclass for all cfn stacks."""

    def __new__(cls, name, bases, attrs):
        super_new = super().__new__

        # Only perform custom logic for the subclasses of Stack, but not Stack itself.
        parents = [b for b in bases if isinstance(b, StackBase)]
        if not parents:
            return super_new(cls, name, bases, attrs)

        new_class = super_new(cls, name, bases, attrs)
        return new_class


class Resource(metaclass=ResourceBase):
    """Represents a resource defintion in Cloudformation."""

    @property
    def logical_name(self):
        """Return the logical of the resource, mapping to the name of the class."""
        return self.__class__.__name__

    @property
    def resource_type(self):
        """Return the type of the resource by analysing the import path."""
        base_class = type(self).__base__
        while True:
            parent = base_class.__base__
            if parent.__module__ == "vapor.models" and parent.__name__ == "Resource":
                break
            base_class = parent
        return f"AWS::{base_class.__module__}::{base_class.__name__}"

    @property
    def template(self):
        """Return the template fragment of the resource."""
        return {
            "Type": self.resource_type,
            "Properties": self.properties,
        }

    @property
    def properties(self):
        """Return the properties of the resource."""
        return {
            name: replace_fn(getattr(self, name))
            for name in dir(self)
            if name[0].isupper()
        }



#!/usr/bin/env python3
"""
Utility functions used in vapor.
"""
import logging
import sys


class ColorFormatter(logging.Formatter):
    """Logging Formatter with color."""

    grey = "\x1b[38;21m"
    yellow = "\x1b[33;21m"
    red = "\x1b[31;21m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    log_format = "[%(asctime)s] %(message)s <%(filename)s:%(lineno)d>"

    FORMATS = {
        logging.DEBUG: grey + log_format + reset,
        logging.INFO: grey + log_format + reset,
        logging.WARNING: yellow + log_format + reset,
        logging.ERROR: red + log_format + reset,
        logging.CRITICAL: bold_red + log_format + reset,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, datefmt="%Y-%m-%d %H:%M:%S")
        return formatter.format(record)


def get_logger(name):
    """create a colorful logger instance."""
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    handler = logging.StreamHandler(stream=sys.stdout)
    handler.setLevel(logging.DEBUG)

    handler.setFormatter(ColorFormatter())
    logger.addHandler(handler)
    return logger



#!/usr/bin/env python3
"""Model definitions in vapor."""
import json
import random
import re
import time
from datetime import datetime

import boto3
import yaml
from botocore.exceptions import ClientError

from .models import StackBase
from .utils import get_logger


logger = get_logger(__name__)


def format_name(name):
    """
    Generate a stack name from class name by converting camel case to dash case.
    Adapted from https://stackoverflow.com/questions/1175208/.

    This function is our best effort to provide a good stack name in Cloudformation.
    However, if you want to have fine-grained control, please specify the name field in
    DeployOptions in your Stack class.

    example: format_name("TestStack") == "test-stack"
    """
    name = re.sub("(.)([A-Z][a-z]+)", r"\1-\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1-\2", name).lower()


def format_changes(changes):
    """
    Format changes so it will look better.

    ref on changeset:

    https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/using-cfn-updating-stacks-changesets-view.html
    """
    parts = []
    for change in changes:
        change_ = change["ResourceChange"]
        line = (
            f"[{change_['Action'].upper()}] "
            f"{change_['LogicalResourceId']}({change_['ResourceType']})"
        )
        if "Details" in change_ and change_["Details"]:
            line += f":\n\t{change_['Details']}"
        parts.append(line)
    return "\n".join(parts)


class Stack(metaclass=StackBase):
    """Represents a Cloudformation stack."""

    def __init__(self):
        self.client = boto3.client("cloudformation")

    @property
    def name(self):
        """Name of the stack."""
        return self.deploy_options.get("name", format_name(self.__class__.__name__))

    @property
    def deploy_options(self):
        """Shortcut to DeployOptions dict provided in Subclasses."""
        return getattr(self, "DeployOptions", {})

    @property
    def template(self):
        """Internal python representation of a Cloudformation template."""
        if not hasattr(self, "Resources"):
            raise ValueError("Please define Resources in your stack.")
        # Resources is defined in child classes.
        # pylint: disable=E1101
        tmplt = {
            "AWSTemplateFormatVersion": "2010-09-09",
            "Resources": {kls().logical_name: kls().template for kls in self.Resources},
        }
        optionals = [
            "Conditions",
            "Mappings",
            "Metadata",
            "Outputs",
            "Parameters",
            "Rules",
            "Transform",
        ]
        for name in optionals:
            if hasattr(self, name):
                tmplt[name] = getattr(self, name)

        return tmplt

    @property
    def json(self):
        """Return Cloudformaiton template in json format"""
        return json.dumps(self.template, indent=2)

    @property
    def yaml(self):
        """Return Cloudformaiton template in yaml format"""
        # Disable anchor in yaml

        # pyyaml will try to be smart and add an anchor to the generated yaml file.
        # This "feature" is not desirable.
        yaml.Dumper.ignore_aliases = lambda self, data: True
        return yaml.dump(self.template, Dumper=yaml.Dumper)

    @property
    def describe_stack_response(self):
        """Return the response from descibe-stacks call."""
        try:
            return self.client.describe_stacks(StackName=self.name)["Stacks"][0]
        except ClientError as error:
            code = error.response["Error"]["Code"]
            message = error.response["Error"]["Message"]
            if code == "ValidationError" and message.endswith("does not exist"):
                return {"StackStatus": "DOES_NOT_EXIST"}
            raise

    @property
    def status(self):
        """Return the status of the stack as a string."""
        return self.describe_stack_response["StackStatus"]

    @property
    def exists(self):
        """Check whether this stack exists."""
        return self.status != "DOES_NOT_EXIST"

    def deploy(self, dryrun=True, wait=True):
        """Wrapper around different steps in the stack deployment process."""
        self.pre_deploy(dryrun, wait)
        self.__deploy(dryrun, wait)
        self.post_deploy(dryrun, wait)

    def __deploy(self, dryrun, wait):
        """Deploy stack changes via changeset."""
        create, name = self.__create_changeset()
        action = "Creating" if create else "Updating"
        logger.info(f"{action} {self.name} stack.")

        changes = self.__wait_changeset(name)
        if changes == []:
            logger.info(f"No change in {self.name} stack.")
            return

        logger.info(f"Changes in changeset `{name}`: \n{format_changes(changes)}")

        if not dryrun:
            logger.info(f"Executing changeset `{name}`.")
            self.client.execute_change_set(ChangeSetName=name, StackName=self.name)
            if wait and not dryrun:
                self.__wait_stack()
        else:
            self.client.delete_change_set(ChangeSetName=name, StackName=self.name)
            if create:
                # An empty stack will stay there until we delete it.
                self.client.delete_stack(StackName=self.name)
            logger.info("Skipping deployment as this is a dry run.")

    def pre_deploy(self, dryrun, wait):
        """Allowing the subclass to add additional steps before the deployment."""
        if self.status == "ROLLBACK_COMPLETE":
            logger.info("Deleting stack in ROLLBACK_COMPLETE state.")
            if not dryrun:
                self.delete(dryrun, wait)

    def post_deploy(self, dryrun, wait):
        """Allowing the subclass to add additional steps after the deployment."""
        # this is a hook, user can potentially use `self` in a subclass.
        # pylint: disable=R0201
        logger.info(f"Nothing to be done in post-deploy hook. {dryrun=}, {wait=}.")

    def delete(self, dryrun=True, wait=True):
        """Wrapper around different steps in the stack deletion process."""
        self.pre_delete(dryrun, wait)
        self.__delete(dryrun, wait)
        self.post_delete(dryrun, wait)

    def pre_delete(self, dryrun, wait):
        """Allowing the subclass to add additional steps before the deletion."""
        # this is a hook, user can potentially use `self` in a subclass.
        # pylint: disable=R0201
        logger.info(f"Nothing to be done in pre-delete hook. {dryrun=}, {wait=}.")

    def post_delete(self, dryrun, wait):
        """Allowing the subclass to add additional steps after the deletion."""
        # this is a hook, user can potentially use `self` in a subclass.
        # pylint: disable=R0201
        logger.info(f"Nothing to be done in post-delete hook. {dryrun=}, {wait=}.")

    def __delete(self, dryrun, wait):
        """Delete stack."""
        if dryrun:
            logger.info("Skipping deletion as this is a dry run.")
        else:
            logger.info(f"Deleting stack: `{self.name}`.")
            self.client.delete_stack(StackName=self.name)
            if wait and not dryrun:
                self.__wait_stack()

    def __create_changeset(self):
        """Create a changeset."""
        suffix = bytearray(random.getrandbits(8) for _ in range(4)).hex()
        name = f"{datetime.now().strftime('%F-%H-%M-%S')}-{suffix}"
        parameters = [
            {"ParameterKey": key, "ParameterValue": value}
            for key, value in self.deploy_options.get("parameters", {}).items()
        ]
        tags = [
            {"Key": tag, "Value": value}
            for tag, value in self.deploy_options.get("tags", {}).items()
        ]
        kwargs = {
            "StackName": self.name,
            "Parameters": parameters,
            "Tags": tags,
            "ChangeSetName": name,
            "ChangeSetType": "UPDATE" if self.exists else "CREATE",
            "Capabilities": self.deploy_options.get("capabilities", []),
            "TemplateBody": self.yaml,
        }

        logger.info(f"Creating changeset {name}.")
        self.client.create_change_set(**kwargs)
        return kwargs["ChangeSetType"] == "CREATE", name

    def __wait_changeset(self, name):
        """Wait till a changeset is available and return it's changes."""
        kwargs = {"ChangeSetName": name, "StackName": self.name}

        while True:
            response = self.client.describe_change_set(**kwargs)
            status = response["Status"]
            exec_status = response["ExecutionStatus"]

            if status == "FAILED":
                empty_changeset_reasons = [
                    "didn't contain changes",
                    "No updates are to be performed.",
                ]
                for reason in empty_changeset_reasons:
                    if reason in response["StatusReason"]:
                        return []
                raise RuntimeError(
                    f"Failed to create changeset for {self.name}: {response['StatusReason']}"
                )

            if status == "CREATE_COMPLETE" and exec_status == "AVAILABLE":
                if not response.get("NextToken"):
                    # Don't have lot's of changes, no need for another api call.
                    return response.get("Changes", [])
                changes = response.get("Changes", [])
                # setting NextToken in kwargs.
                kwargs["NextToken"] = response["NextToken"]
                break
            logger.info(
                f"Status of changeset is `{status}`, execution status is `{exec_status}`"
            )
            # Change set should be ready within seconds.
            time.sleep(3)

        while True:
            response = self.client.describe_change_set(**kwargs)
            changes += response.get("Changes", [])
            if not response.get("NextToken"):
                break
            kwargs["NextToken"] = response["NextToken"]
        return changes

    def __wait_stack(self):
        while True:
            current_status = self.status
            if current_status == "DOES_NOT_EXIST":
                # Delete complete
                return

            if current_status.split("_")[-1] in ["FAILED", "COMPLETE"]:
                break
            logger.info(f"Waiting till stack operation completes: {current_status}.")
            time.sleep(5)

        logger.info(f"Stack operation finished: {current_status}")
        bad_statuses = [
            "UPDATE_ROLLBACK_COMPLETE",
            "ROLLBACK_COMPLETE",
        ]
        if current_status in bad_statuses or current_status.endswith("FAILED"):
            raise RuntimeError("Failed to create/update stack.")



#!/usr/bin/env python3
"""
finder to create dynamic modules.
"""
import sys

from importlib.machinery import ModuleSpec

from .models import ResourceBase, StackBase, Resource
from .fn import Ref, Fn
from .stack import Stack


class AWSFinder:
    """
    This class Follows the python module loading protocol and will serve as a virtual
    module.

    Ref: PEP-302
    """

    @classmethod
    def find_spec(cls, name, path, target=None):
        """
        If we are importing vapor things, create a module for it.
        """
        # This is just following protocol
        # pylint: disable=W0613
        if name.startswith("vapor."):
            _, service = name.split(".", 1)
            return ModuleSpec(service, cls())
        return None

    def create_module(self, _):
        """Do nothing specific here."""
        # This is just following protocol
        # pylint: disable=R0201
        return None

    def exec_module(self, module):
        """Create a dynamic class and return it."""
        # This is just following protocol
        # pylint: disable=R0201
        def _getattr(name):
            cls = type(name, (Resource,), {"_module": module})
            return cls

        module.__getattr__ = _getattr


sys.meta_path += [AWSFinder]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vapor']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0', 'boto3>=1.18,<2.0']

setup_kwargs = {
    'name': 'vapor',
    'version': '0.0.5',
    'description': 'Django ORM meets Cloudformation',
    'long_description': '# vapor\n\n![build](https://github.com/xiaket/vapor/workflows/build/badge.svg)\n![PyPI version](https://badge.fury.io/py/vapor.svg)\n![Coverage](https://coveralls.io/repos/github/xiaket/vapor/badge.svg)\n![license](https://img.shields.io/pypi/l/vapor)\n\nVapor is an alternative way to write and manage your cloudformation stacks. Think of CDK, but with a promise to generate a template that is easy to understand and easy to manage. Also, if you are familiar with Django ORM, the way we define resources is really similar to how we define models in Django.\n',
    'author': 'Kai Xia',
    'author_email': 'kaix+github@fastmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/xiaket/vapor',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
